###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
Implements classes and routines for manipulating Metadata and interacting
with files and file-like data in Coscine.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, List, Callable
if TYPE_CHECKING:
	from .client import Client
	from .resource import Resource
from .defaults import TABLE_FORMATS
from .exceptions import *
from .logger import ProgressBar
from .graph import ApplicationProfile
from .form import InputForm, FormField
from .utils import HumanBytes
from prettytable.prettytable import PrettyTable
import os

###############################################################################
# Class definition
###############################################################################

class MetadataForm(InputForm):
	"""
	The MetadataForm supports parsing coscine.Object metadata, generating
	metadata and manipulating it in the same way one would manipulate
	a python dict.
	"""

	profile: ApplicationProfile

###############################################################################

	def __init__(self, client: Client, graph: ApplicationProfile) -> None:
		"""
		Initializes an instance of type MetadataForm.

		Parameters
		----------
		client : Client
			Coscine Python SDK client handle
		graph : str
			Coscine application profile rdf graph as json-ld string
		"""

		self.profile = graph
		super().__init__(client)
		self._items = [FormField(client, item) for item in self.profile.items()]
		# Query vocabularies for controlled fields
		for item in self._items:
			if item.vocabulary:
				self._vocabularies[item.path] = \
					client.vocabularies.instance(item.vocabulary)

###############################################################################

	def generate(self) -> dict:
		"""
		Generates and validates metadata for sending to Coscine.
		"""

		metadata = {}

		metadata[ApplicationProfile.RDFTYPE] = [{
			"type": "uri",
			"value": self.profile.target()
		}]

		# Collect missing required fields
		missing = []

		# Set metadata
		for key, values in self.items(True):
			
			if not values:
				if self.is_required(key):
					missing.append(key)
				continue

			properties = self.properties(key)
			metadata[properties.path] = []
			for value in values:
				entry = {
					"value": value,
					"datatype": properties.datatype \
						if properties.datatype else properties.vocabulary,
					"type": "uri" if properties.vocabulary else "literal"
				}
				if not entry["datatype"]:
					del entry["datatype"]
				metadata[properties.path].append(entry)

		# Check for missing required fields
		if len(missing) > 0:
			raise ValueError(missing)

		return metadata

###############################################################################

	def parse(self, data: dict) -> None:
		if data is None:
			return
		for path, values in data.items():
			if path == ApplicationProfile.RDFTYPE:
				continue
			key = self.name_of(path)
			self.set_value(key, [v["value"] for v in values], True)

###############################################################################
# Class definition
###############################################################################

class FileObject:
	"""
	Objects in Coscine represent file-like data. We could have called it
	'File', but in case of linked data we are not actually dealing with files
	themselves, but with links to files.Thus we require a more general
	datatype.
	"""

	client: Client
	resource: Resource
	data: dict

	name: str
	size: int
	type: str
	path: str
	is_folder: bool

	@property
	def name(self) -> str: return self.data["Name"]

	@property
	def size(self) -> int: return int(self.data["Size"])

	@property
	def type(self) -> str: return self.data["Kind"]

	@property
	def filetype(self) -> str:
		return os.path.splitext(self.data["Name"])[1]

	@property
	def path(self) -> str: return self.data["Path"]

	@property
	def is_folder(self) -> bool: return bool(self.data["IsFolder"])

	CHUNK_SIZE: int = 4096

###############################################################################

	def __init__(self, resource: Resource, data: dict) -> None:
		"""
		Initializes the Coscine FileObject.

		Parameters
		----------
		resource : Resource
			Coscine resource handle.
		data : dict
			data of the file-like object.
		"""

		# Configuration
		self.client = resource.client
		self.resource = resource
		self.data = data

###############################################################################

	def __str__(self) -> str: return self.str()

###############################################################################

	def str(self, format: str = "str") -> str:
		"""
		Returns a string in the specified format

		Parameters
		-----------
		format : str, default "str"
			Format of the string, possible options are: str, csv, json, html

		Returns
		--------
		str
		"""

		if format not in TABLE_FORMATS:
			raise ValueError(f"Unsupported format '{format}'!")

		table = PrettyTable(("Property", "Value"))
		rows = [
			("Name", self.name),
			("Size", HumanBytes.format(self.size)),
			("Type", self.type),
			("Path", self.path),
			("Folder", self.is_folder)
		]
		table.max_width["Value"] = 50
		table.add_rows(rows)
		if format == "str":
			return table.get_string(title = "Object [%s]" % self.name)
		elif format == "csv":
			return table.get_csv_string(title = "Object [%s]" % self.name)
		elif format == "json":
			return table.get_json_string(title = "Object [%s]" % self.name)
		elif format == "html":
			return table.get_html_string(title = "Object [%s]" % self.name)

###############################################################################

	def metadata(self) -> dict:
		"""
		Retrieves the metadata of the file-like object.

		Returns
		-------
		dict
			Metadata as a python dictionary.
		None
			If no metadata has been set for the file-like object.
		"""

		uri = self.client.uri("Tree", "Tree", self.resource.id)
		args = {"path": self.path}
		data = self.client.get(uri, params=args).json()
		metadata = data["data"]["metadataStorage"]
		if not metadata:
			return None
		metadata = metadata[0]
		for key in metadata:
			return metadata[key]

###############################################################################

	def update(self, metadata) -> None:
		"""
		Updates the metadata of the file-like object.

		Parameters
		----------
		metadata : MetadataForm or dict
			MetadataForm or JSON-LD formatted metadata dict
		
		Raises
		------
		TypeError
			If argument `metadata` has an unexpected type.
		"""

		if type(metadata) is MetadataForm:
			metadata = metadata.generate()
		elif type(metadata) is not dict:
			raise TypeError("Expected MetadataForm or dict.")
		
		uri = self.client.uri("Tree", "Tree", self.resource.id)
		args = {"path": self.path}
		self.client.put(uri, params = args, data = metadata)

###############################################################################

	def content(self) -> bytes:
		"""
		Retrieves the content/data of the object. In case of linked data
		this would be the link, not the actual file itself. It is impossible
		to get the file contents of linked data objects with this python module.
		In case of rds or rds-s3 data this would return the file contents.
		Be aware that for very large files this will consume a considerable
		amount of RAM!

		Returns
		-------
		bytes
			A raw byte-array containing the Coscine file-object's data.
		"""

		uri = self.client.uri("Blob", "Blob", self.resource.id)
		args = {"path": self.path}
		return self.client.get(uri, params = args).content

###############################################################################

	def download(self, path: str = "./", callback: Callable[[int]] = None) -> None:
		"""
		Downloads the file-like object to the local harddrive.

		Parameters
		----------
		path : str, default: "./"
			The path to the download location on the harddrive.
		callback : function(chunksize: int)
			A callback function to be called during downloading chunks.
		"""

		uri = self.client.uri("Blob", "Blob", self.resource.id)
		args = {"path": self.path}
		response = self.client.get(uri, params = args, stream = True)
		path = os.path.join(path, self.name)
		with open(path, 'wb') as fd:
			bar = ProgressBar(self.client.logger, self.size, self.name, "DOWN", callback)
			for chunk in response.iter_content(chunk_size = self.CHUNK_SIZE):
				fd.write(chunk)
				bar.update(len(chunk))

###############################################################################

	def delete(self) -> None:
		"""
		Deletes the file-like object on the Coscine server.
		"""

		uri = self.client.uri("Blob", "Blob", self.resource.id)
		args = {"path": self.path}
		self.client.delete(uri, params = args)

###############################################################################

	def objects(self, **kwargs) -> List[FileObject]:
		if self.is_folder:
			return self.resource.objects(path=self.path, **kwargs)
		else:
			raise TypeError("object is not a directory!")

###############################################################################

	def object(self, displayName: str = None, **kwargs) -> FileObject:
		if self.is_folder:
			if displayName.endswith("/"): # expect directory
				displayName = self.name + "/" + displayName
			return self.resource.object(displayName = displayName,
											path=self.path, **kwargs)
		else:
			raise TypeError("object is not a directory!")

###############################################################################

	def form(self) -> MetadataForm:
		"""
		Returns a MetadataForm to interact with the metadata of the FileObject.
		"""

		graph = self.resource.application_profile()
		form = MetadataForm(self.client, graph)
		form.parse(self.metadata())
		return form

###############################################################################

	def move(self, resource: Resource) -> None:
		"""
		Move a file-like object to another resource
		"""
		resource.upload(self.name, self.content(), self.metadata())

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file provides base class for all input forms defined by
the Coscine Python SDK.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, Union, MutableMapping, List, Tuple, Iterator
if TYPE_CHECKING:
	from .client import Client
	from .vocabulary import Vocabulary
from prettytable import PrettyTable
from datetime import datetime
from enum import Enum
from .exceptions import *
from .defaults import TIMEFORMAT, LANGUAGES

###############################################################################
# Class
###############################################################################

class FormField:
	"""
	The FormField class defines a field such as "Display Name"
	inside an InputForm with all of its properties.
	"""

	client: Client
	_data: dict

	def __init__(self, client: Client, data: dict) -> None:
		self.client = client
		self._data = data
	
	@property
	def name(self) -> str:
		return self._data["name"][LANGUAGES.index(self.client.language)]
	
	@property
	def path(self) -> str: return self._data["path"]

	@property
	def order(self) -> int: return int(self._data["order"])

	@property
	def vocabulary(self) -> str: return self._data["class"]

	@property
	def minCount(self) -> int: return int(self._data["minCount"])

	@property
	def maxCount(self) -> int: return int(self._data["maxCount"])

	@property
	def datatype(self) -> str: return self._data["datatype"]

###############################################################################
# Class
###############################################################################

class ValueFormat(Enum):
	"""
	Defines the value format

	Attributes
	-----------
	EXTERNAL
		The value format as apparent to the user (e.g. Informatik 409)
	INTERNAL
		The value format expected by Coscine (e.g. dict{diplayName: ...})
	UNIQUE
		Unique value identifier
	"""
	EXTERNAL = 0
	INTERNAL = 1
	UNIQUE = 2

###############################################################################
# Class
###############################################################################

class InputForm(MutableMapping):
	"""
	Coscine InputForm base class
	"""

	client: Client

###############################################################################

	def __init__(self, client: Client) -> None:
		super().__init__()
		self.client = client
		self._items: List[FormField]
		self._values: dict
		self._vocabularies: dict[str, Vocabulary]

###############################################################################

	def __setitem__(self, key: str, value: object) -> None:
		self.set_value(key, value)

###############################################################################

	def __getitem__(self, key: str) -> Union(List, str):
		return self.get_value(key)

###############################################################################

	def __delitem__(self, key: str) -> None: del self._values[self.path(key)]

###############################################################################

	def __iter__(self) -> Iterator[str]: return iter(self.keys())

###############################################################################

	def __len__(self) -> int: return len(self._items)

###############################################################################

	def __str__(self) -> str: return self.str()

###############################################################################

	def clear(self) -> None:
		"""Remove all values of the InputForm"""
		self._values.clear()

###############################################################################

	def path(self, key: str) -> str:
		"""
		Returns the unique identifier for a multilanguage key.
		Logical example:
		path("Display Name") == path("Anzeigename") == "__XXX__"
		->> True
		where "__XXX__" is of some fixed internal value

		Parameters
		-----------
		key : str
			Multilanguage key
		"""
		return self.properties(key).path

###############################################################################

	def properties(self, key: str) -> FormField:
		"""
		Returns the form field properties corresponding to the given key.
		"""
		return self._items[self.index_of(key)]

###############################################################################

	def keys(self) -> List[str]:
		"""
		Returns a list of keys in their respective order based on the
		language setting of the client class instance used to initialize
		the InputForm.
		"""
		return [item.name for item in self._items]

###############################################################################

	def values(self, raw: bool = False) -> List[Union[str, List[str]]]:
		"""
		Returns a list of values in their respective order based on the
		language setting of the client class instance used to initialize
		the InputForm.

		Parameters
		----------
		raw : bool, default: False
			If set to true, values are not returned in a user-friendly
			format but instead in the actual language-independent format.
		"""
		return [self.get_value(key, raw) for key in self.keys()]

###############################################################################

	def items(self, raw: bool = False) -> List[Tuple[str, Union[str, List]]]:
		"""
		Returns a list of key, value pairs in their respective order based
		on the 	language setting of the client class instance
		used to initialize the InputForm.

		Parameters
		----------
		raw : bool, default: False
			If set to true, values are not returned in a user-friendly
			format but instead in the actual language-independent format.
		"""
		return zip(self.keys(), self.values(raw))

###############################################################################

	def is_valid(self, key: str, value: str, raw: bool = False) -> bool:
		"""
		Determines whether a given value for a key is valid based on
		the language setting of the client and if the value is part
		of a vocabulary in case of controlled fields.
		"""
		if self.is_required(key) and not value:
			raise ValueError(f"Expected value for required field {key}!")
		if not value: return True
		elif type(value) is not str and type(value) is not dict \
			 				and type(value) is not self.datatype(key):
			raise TypeError(
				f"Value set for key '{key}' is of invalid type:\n"\
				f"Expected str or {str(self.datatype(key))} but got "\
				f"{str(type(value))} instead."
			)
		if not self.is_controlled(key): return True
		else: return self.vocabulary(key).contains(value, raw)

###############################################################################

	def is_required(self, key: str) -> bool:
		"""
		Determines whether a key is a required one.
		"""
		return self.properties(key).minCount > 0

###############################################################################

	def is_controlled(self, key: str) -> bool:
		"""
		Determines whether a key is a controlled one.
		"""
		return self.properties(key).vocabulary is not None

###############################################################################

	def vocabulary(self, key: str) -> Vocabulary:
		"""
		Returns the vocabulary for the given key.
		"""
		if self.is_controlled(key):
			return self._vocabularies[self.path(key)]
		else:
			raise KeyError("Key is not controlled by a vocabulary!")

###############################################################################

	def index_of(self, key: str) -> int:
		"""
		Returns the order of the key.
		"""
		for index, item in enumerate(self.keys()):
			if item == key:
				return index
		raise KeyError(f"Key '{key}' not found in InputForm!")

###############################################################################

	def name_of(self, path: str) -> str:
		"""
		Returns the key name for the unique path
		"""
		for item in self._items:
			if item.path == path:
				return item.name
		raise KeyError(f"Path '{path}' not found!")

###############################################################################

	def datatype(self, key: str) -> type:
		"""
		Returns the datatype
		"""
		datatype = self.properties(key).datatype
		if datatype == "xsd:string":
			return str
		elif datatype == "xsd:boolean":
			return bool
		elif datatype == "xsd:date":
			return datetime
		else:
			return None

###############################################################################

	def is_typed(self, key: str) -> bool:
		"""
		Returns whether a value must have a certain datatype
		"""
		return self.datatype(key) is not None

###############################################################################

	def parse_value(self, key: str, value: object, reverse: bool = False) -> object:
		"""
		Converts a value to the datatype specified in its field.
		e.g. for a timestring "2022-01-01T10:12:00" returns a datetime object.
		"""

		if self.is_typed(key) and type(value) is str:
			if self.datatype(key) == datetime:
				return datetime.strptime(value, TIMEFORMAT)
		if self.is_controlled(key):
			value = self.vocabulary(key).lookup(value, reverse)

		return value

###############################################################################

	def add_value(self, key: str, value: object, raw: bool = False) -> None:
		if not self.is_valid(key, value, raw):
			raise ValueError(f"Invalid value for key '{key}': {value}")
		value = self.parse_value(key, value, raw)
		if not self.path(key) in self._values:
			self._values[self.path(key)] = [value]
		else:
			self._values[self.path(key)].append(value)

###############################################################################

	def set_value(self, key: str, value: object, raw: bool = False) -> None:
		if type(value) is list:
			for val in value:
				self.add_value(key, val, raw)
		else:
			if not self.is_valid(key, value, raw):
				raise ValueError(f"Invalid value for key '{key}': {value}")
			value = self.parse_value(key, value)
			self._values[self.path(key)] = [value]

###############################################################################

	def get_value(self, key: str, raw: bool = False) -> Union(List, str):
		path = self.path(key)
		if path in self._values:
			if self.is_controlled(key):
				if raw: return self._values[path]
				else: return [self.vocabulary(key).lookup(v, True) \
										for v in self._values[path]]
			else:
				return self._values[path]
		return []

###############################################################################

	def parse(self, data: dict) -> None:
		"""
		Parses data from Coscine into an InputForm.
		"""

###############################################################################

	def generate(self) -> dict:
		"""
		Generates Coscine formatted json-ld from an InputForm.
		"""

###############################################################################

	def value_str(self, key: str, raw: bool = False) -> str:
		"""
		Returns a value as a string
		"""
		string: str = ""
		values = self.get_value(key, raw)
		if len(values) > 1:
			string = "\n".join(values)
		elif len(values) == 1:
			string = values[0] if values[0] else ""
		return string

###############################################################################

	def str(self, format: str = "str", raw: bool = False) -> str:
		"""
		Returns a string in the specified format

		Parameters
		-----------
		format : str, default "str"
			Format of the string, possible options are: str, csv, json, html
		raw : bool, default False
			If set to true, values of the form are in Coscine internal format
		
		Returns
		--------
		str
		"""

		SUPPORTED_FORMATS = ["str", "csv", "json", "html"]
		if format not in SUPPORTED_FORMATS:
			raise ValueError(f"Unsupported format '{format}'!")
		table = PrettyTable(("C", "Property", "Value"), align="l")
		rows = []
		for key in self.keys():
			value = self.value_str(key, raw)
			name: str = key
			name = name + "*" if self.is_required(key) else name
			c: str = "X" if self.is_controlled(key) else ""
			rows.append((c, name, value))
		table.max_width["Value"] = 50
		table.add_rows(rows)
		if format == "str":
			return table.get_string()
		elif format == "csv":
			return table.get_csv_string()
		elif format == "json":
			return table.get_json_string()
		elif format == "html":
			return table.get_html_string()

###############################################################################



###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file provides a simple wrapper around Coscine application profiles.
It abstracts the interaction with rdf graphs using rdflib and
provides an intuitive, Coscine-specific interface to these graphs.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import List
import rdflib

###############################################################################
# Class
###############################################################################

class ApplicationProfile:
	"""
	The ApplicationProfile class serves as a wrapper around Coscine
	application profiles. Coscine application profiles are served
	as rdf graphs which are difficult to interact with, without 3rd
	party libraries.
	The ApplicationProfile class abstracts this interaction with rdf
	graphs.

	Attributes
	----------
	graph : rdflib.Graph
		An rdf graph parsed with the help of rdflib
	RDFTYPE : str
		A variable frequently used by methods of this class.
	"""

	graph: rdflib.Graph
	RDFTYPE: str = "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"

###############################################################################

	def __init__(self, graph: str, format: str = "json-ld") -> None:
		"""
		Initializes an instance of the ApplicationProfile wrapper class.

		Parameters
		----------
		graph : str
			A Coscine application profile rdf graph in json-ld text format.
		format : str (optional), default: "json-ld"
			The format of the graph
		"""

		self.graph = rdflib.Graph()
		self.graph.bind("sh", "http://www.w3.org/ns/shacl#")
		self.graph.bind("dcterms", "http://purl.org/dc/terms/")
		self.graph.parse(data = graph, format = format)

###############################################################################

	def __str__(self) -> str:
		"""
		Serializes the application profile rdf graph used internally
		for easy output to stdout.
		"""
		return self.graph.serialize(format="ttl")

###############################################################################

	def target(self) -> str:
		"""
		Returns a str indicating the target class of the application profile.
		This may for example be "engmeta" in case of an engmeta profile.
		"""

		QUERY = \
			"""
			SELECT ?target WHERE {
				?_ sh:targetClass ?target .
			}
			"""
		result = self.query(QUERY)
		return result[0][0]

###############################################################################

	def query(self, query: str, **kwargs) -> List[List[object]]:
		"""
		Performs a SPARQL query against the application profile and
		returns the result as a list of rows.
		"""

		items = []
		results = self.graph.query(query, **kwargs)
		for row in results:
			item = []
			for val in row:
				value = val.toPython() if val is not None else None
				item.append(value)
			items.append(item)
		return items

###############################################################################

	def items(self) -> List[dict]:
		"""
		Returns all items contained within the application profile as a list
		of key value pairs in their order of appearance.
		"""

		QUERY = \
			"""
			SELECT ?path ?name ?order ?class ?minCount ?maxCount ?datatype (lang(?name) as ?lang) WHERE {
				?_ sh:path ?path ;
					sh:name ?name ;
					sh:order ?order .
				OPTIONAL { ?_ sh:class ?class . } .
				OPTIONAL { ?_ sh:minCount ?minCount . } .
				OPTIONAL { ?_ sh:maxCount ?maxCount . } .
				OPTIONAL { ?_ sh:datatype ?datatype . } .
			}
			ORDER BY ASC(?order)
			"""
		items_de = []
		items_en = []
		items = []
		results = self.query(QUERY)
		for result in results:
			lang = result[7]
			properties = {
				"path": result[0],
				"name": result[1],
				"order": result[2],
				"class": result[3],
				"minCount": result[4] if result[4] else 0,
				"maxCount": result[5] if result[5] else 42,
				"datatype": result[6]
			}
			if lang == "de": items_de.append(properties)
			else: items_en.append(properties)

		# Fuse language lists into one
		for index, it in enumerate(items_de):
			it["name"] = [it["name"], items_en[index]["name"]]
			items.append(it)
		return items

###############################################################################

	def length(self) -> int:
		"""
		Returns the number of fields contained within the application profile.
		"""

		QUERY = \
			"""
			SELECT ?path WHERE {
				?_ sh:path ?path ;
					sh:order ?order .
			}
			"""
		results = self.query(QUERY)
		return len(results)

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file contains the backbone of the Coscine Python SDK - the client class.
The client class acts as the manager of the SDK and is mainly
responsible for the communication and exchange of information 
with Coscine servers.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import List
import sys
import json
from json.decoder import JSONDecodeError
import requests
import urllib.parse
from .logger import Logger
from .cache import Cache
from .vocabulary import VocabularyManager
from .graph import ApplicationProfile
from .project import Project, ProjectForm
from .resource import ResourceForm
from .object import FileObject, MetadataForm
from .__about__ import __version__
from .defaults import LANGUAGES
from .exceptions import *

###############################################################################
# Class
###############################################################################

class FormManager:
	"""
	Serves empty InputForms
	"""

	client: Client

	def __init__(self, client: Client) -> None:
		self.client = client
	def metadata(self, graph: ApplicationProfile) -> MetadataForm:
		"""
		Generates a metadata form
		"""
		return MetadataForm(self.client, graph)
	def project(self, parent: Project = None) -> ProjectForm:
		"""
		Generates a project form
		"""
		return ProjectForm(self.client, parent)
	def resource(self) -> ResourceForm:
		"""
		Generates a resource form
		"""
		return ResourceForm(self.client)

###############################################################################
# Class
###############################################################################

class Client:
	"""
	The client class is the backbone and manager of the SDK and mainly
	responsible for the communication and exchange of information 
	with Coscine servers.

	Attributes
	-----------
	language : str, "en" or "de"
		Language preset for input forms and form data.
	version : str
		Contains the current version string of the python SDK.
	logger : Logger
		An instance of the coscine.Logger class for cli/file output.
	session : requests.Session
		A requests session for the communication with Coscine servers.
	vocabularies : VocabularyManager
		Local cache server for nonvolatile data from Coscine.
	forms : FormManager (Use as FormManager.project/resource/metadata)
		Serves empty InputForms
	"""

	_language: str
	logger: Logger
	cache: Cache
	vocabularies: VocabularyManager
	forms: FormManager
	session: requests.Session
	concurrent: bool

	@property
	def version(self) -> str: return __version__

	@property
	def language(self) -> str: return self._language

	@language.setter
	def language(self, lang: str) -> None:
		lang = lang.lower()
		if lang in LANGUAGES:
			self._language = lang
		else:
			raise ValueError(f"Invalid value for argument 'lang' -> [{lang}]!\n"
									f"Possible values are {str(LANGUAGES)}.")

###############################################################################

	def __init__(self, token: str, lang: str = "en", logger: Logger = None,
			persistent_cache: bool = True, concurrent: bool = True) -> None:
		"""
		Initializes an instance of the base class of the Coscine Python SDK.

		Parameters
		----------
		token : str
			A Coscine API access token.
		lang : str, "en" or "de", default: "en"
			Language preset for input form fields and vocabularies.
		logger : Logger
			Sets a logger object for command line or file output.
		persistent_caching : bool, default: True
			Enable to store the cache in a file on deinitialization of
			the client object. Will attempt to load the cache file
			on initialization if enabled. Leads to a significant speed
			boost when making static requests right after init, but may
			also lead to invalid data when using outdated cache data, in case
			the Coscine API changed recently. However this is mostly avoided
			by performing frequent updates. Useful for applications with
			a short runtime that get run often.
		concurrent : bool, default: True
			If enabled, a ThreadPool is used for bulk requests, speeding up
			up- and downloads of multiple files tremendously.
		"""

		self.language = lang
		self.concurrent = concurrent
		enableLogging = False
		# If we are running from a command line
		if sys.stdout != None and sys.stderr.isatty():
			enableLogging = True
		self.logger = logger if logger else Logger(enableLogging, sys.stdout)
		self.cache = Cache(persistent_cache)
		self.vocabularies = VocabularyManager(self)
		self.forms = FormManager(self)
		self.session = requests.Session()
		self.session.headers = {
			"Authorization": f"Bearer {token}",
			"User-Agent": f"Coscine Python SDK {self.version}"
		}
		self.logger.banner()

###############################################################################

	@staticmethod
	def uri(api: str, endpoint: str, *args) -> str:
		"""
		Constructs a URL for performing a request to the Coscine API.

		Parameters
		----------
		api : str
			The target Coscine API endpoint, e.g. Blob, Metadata, Tree, ...
		endpoint : str
			The subendpoint of `api`.
		*args
			Variable number of arguments of type string to append to the URL.
			Arguments are automatically seperated by a slash '/' and
			special characters are encoded.
		
		Raises
		------
		ValueError
			If an invalid value is given as an argument to the method.

		Returns
		-------
		str
			Encoded URL for communication with the Coscine servers.
		"""

		BASE = "https://coscine.rwth-aachen.de/coscine/api/Coscine.Api.%s/%s"
		ENDPOINTS = (
			"Blob",
			"Metadata",
			"Organization",
			"Project",
			"Resources",
			"Tree",
			"User",
			"Search",
			"ActivatedFeatures"
		)

		if api not in ENDPOINTS:
			raise ValueError("Invalid value for argument 'api'!\n" \
						"Possible values are %s." % str(ENDPOINTS))
		
		uri = BASE % (api, endpoint)
		for arg in args:
			if arg is None:
				continue
			uri += "/" + urllib.parse.quote(arg, safe="")
		return uri

###############################################################################

	def _request(self, method: str, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a HTTP request to the Coscine Servers.

		Parameters
		----------
		method : str
			HTTP request method (GET, PUT, POST, DELETE).
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional keyword arguments forwarded to the requests library.
		
		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		# Normalize data before sending
		if "data" in kwargs and type(kwargs["data"]) is dict:
			kwargs["headers"] = {
				"Content-Type": "application/json;charset=utf-8"
			}
			kwargs["data"] = json.dumps(kwargs["data"])

		try:
			self.logger.net(f"{method} {uri}\n{json.dumps(kwargs, indent=4)}")
		except (JSONDecodeError, TypeError):
			pass

		# Perform the request and handle any resulting errors
		try:
			response = self.session.request(method, uri, **kwargs)
			response.raise_for_status()
			self.logger.net(f"response: {response.content}")
			return response
		except requests.exceptions.ConnectionError:
			raise ConnectionError
		except requests.exceptions.RequestException as e:
			if e.response.status_code == 401:
				raise AuthorizationError("Invalid Coscine API token!")
			else:
				raise ClientError()

###############################################################################

	def get(self, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a GET request to the Coscine API.

		Parameters
		----------
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional keyword arguments forwarded to the requests library.
		
		Examples
		--------
		>>> uri = client.uri("Project", "Project")
		>>> projects = client.get(uri).json()

		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		return self._request("GET", uri, **kwargs)

###############################################################################

	def put(self, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a PUT request to the Coscine API.

		Parameters
		----------
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional keyword arguments forwarded to the requests library.
		
		Examples
		--------
		>>> uri = self.uri("Tree", "Tree", resource.id, filename)
		>>> self.put(uri, data = metadata)

		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		return self._request("PUT", uri, **kwargs)

###############################################################################

	def post(self, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a POST request to the Coscine API.

		Parameters
		----------
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional arguments forwarded to the requests library.
		
		Examples
		--------
		>>> data = member.data
		>>> data["projectId"] = self.id
		>>> data["role"]["displayName"] = role
		>>> data["role"]["id"] = ProjectMember.ROLES[role]
		>>> uri = self.uri("Project", "ProjectRole")
		>>> self.post(uri, data = data)

		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		return self._request("POST", uri, **kwargs)

###############################################################################

	def delete(self, uri: str, **kwargs) -> requests.Response:
		"""
		Performs a DELETE request to the Coscine API.

		Parameters
		----------
		uri : str
			Coscine URL generated with Client.uri(...).
		**kwargs
			Additional keyword arguments forwarded to the requests library.
		
		Examples
		--------
		>>> uri = self.uri("Project", "Project", self.id)
		>>> self.delete(uri)

		Raises
		------
		coscine.exceptions.ConnectionError
			If the Coscine servers could not be reached.
		coscine.exceptions.AuthorizationError
			If the Coscine API token is invalid.
		coscine.exceptions.ClientError
			If the request resulted in an error.

		Returns
		-------
		requests.Response
			The response of the Coscine server as a requests.Response object.
		"""

		return self._request("DELETE", uri, **kwargs)

###############################################################################

	def static_request(self, uri: str) -> dict:
		"""
		Performs a GET request for the given uri. If such a request
		has been performed previously during the session, the previous
		response is returned. Otherwise a new request is made to Coscine
		and that response is then stored inside the session cache.

		Parameters
		-----------
		uri : str
			Request URI
		
		Returns
		-------
		dict
		"""

		data = self.cache.get(uri)
		if not data:
			data = self.get(uri).json()
			self.cache.set(uri, data)
		return data

###############################################################################

	def projects(self, toplevel: bool = True, **kwargs) -> List[Project]:
		"""
		Retrieves a list of a all projects the creator of the Coscine API token
		is currently a member of.

		Parameters
		----------
		toplevel : bool, default: True
			Retrieve only toplevel projects (no subprojects).
			Set to False if you want to retrieve all projects, regardless
			of hierarchy.
		**kwargs
			Project filter values.
			-> e.g. displayName="MyProject"
			-> Returns all projects with the specified displayName

		Returns
		-------
		list
			List of coscine.Project objects
		"""

		ENDPOINTS = ("Project", "Project/-/topLevel")
		uri = self.uri("Project", ENDPOINTS[toplevel])
		projects = []
		for it in self.get(uri).json():
			for key, value in kwargs.items():
				if it[key] != value:
					break
			else:
				projects.append(Project(self, it))
		return projects

###############################################################################

	def project(self, displayName: str = None, toplevel: bool = True,
												**kwargs) -> Project:
		"""
		Retrieves a single project the owner of the Coscine
		API token is currently a member of. Search criteria
		must be specified.

		Parameters
		----------
		displayName : str, default: None
			Default displayName search argument. May be omitted
			in favor of different arguments given in $kwargs. Allows
			easier use as `client.project("name")`.
		toplevel : bool, default: True
			Search only within toplevel projects (no subprojects)
			Set to False if you want to search through all projects,
			regardless of hierarchy.
		**kwargs (mandatory!)
			Project filter values.
			-> e.g. displayName="MyProject"
			-> Returns one project matching the specified displayName

		Raises
		-------
		OverflowError
			If more than one project match the specified search criteria.

		Returns
		-------
		Project
			coscine.Project object if a project matches the specifed filter.
		None
			if no project matching the filter was found.
		"""

		# Handle Arguments
		if displayName:
			kwargs["displayName"] = displayName
		elif not kwargs:
			raise ValueError("Not enough arguments!")

		# Filter projects
		projects = self.projects(toplevel, **kwargs)
		if len(projects) == 1:
			return projects[0]
		elif len(projects) == 0:
			return None
		else:
			raise OverflowError("More than 1 project matching criteria!")

###############################################################################

	def create_project(self, form: ProjectForm) -> Project:
		"""
		Creates a project using the given ProjectForm.

		Parameters
		----------
		form : ProjectForm
			ProjectForm filled with project metadata.

		Returns
		-------
		Project
			Project object of the new project.
		"""

		if type(form) is ProjectForm:
			form = form.generate()

		uri = self.uri("Project", "Project")
		return Project(self, self.post(uri, data=form).json())

###############################################################################

	def object(self, path: str) -> FileObject:
		"""
		Returns an object for a path

		Parameters
		----------
		path : str
			In format of "Project/Resource/filepath.filetype"

		Returns
		-------
		FileObject
		"""

		parts = path.split("/")
		if len(parts) < 3:
			raise ValueError("Not enough components in path. Must be > 2.")
		filepath = "/".join(parts[2:])
		return self.project(parts[0]).resource(parts[1]).object(filepath)

###############################################################################

	def search(self, query: str) -> dict:
		"""
		Performs a search query on Coscine
		"""

		uri = self.uri("Search", f"SemanticSearch?query={query}")
		results = self.get(uri).json()
		return results

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file provides an easy way of reading coscine-python-sdk
config files.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from .__about__ import __version__
from .defaults import APPDIR
import json
import os

###############################################################################
# Class
###############################################################################

class Config:
	"""
	A class to read from coscine-python-sdk config files.
	Lookup order:  
		0. Custom path specified in constructor (fail if not found)  
		1. Local config file  
		2. Local config directory  
		3. System directory (e.g. User/AppData/Local/... on Windows)  
	"""

	token: str
	language: str
	concurrent: bool

###############################################################################

	def __init__(self, path: str = "") -> None:
		"""
		Initialized the configuration file interface with an
		optional path argument.
		
		Parameters
		-----------
		path : str
			If set to a valid file path, an attempt is made to laod
			the file referenced by path as a config.
		
		Raises
		-------
		FileNotFoundError
			In case no config file could have been located.
		"""

		if path:
			self.load(path)
		elif os.path.exists(".config.json"):
			self.load(".config.json")
		elif os.path.exists(".coscine/"):
			self.load(os.path.join(".coscine/", "config.json"))
		elif os.path.exists(APPDIR.user_config_dir):
			self.load(os.path.join(APPDIR.user_config_dir, "config.json"))
		else:
			raise FileNotFoundError("Failed to locate config file!")

###############################################################################

	def load(self, path: str) -> None:
		"""
		Loads a valid coscine-python-sdk config file and validates
		the file for correctness.

		Parameters
		-----------
		path : str
			The filepath to the configuration file (must include filename)
		"""

		with open(path, "r") as fd:
			data = json.load(fd)
		self.token = data["token"] if "token" in data else None
		self.language = data["language"] if "language" in data else None
		self.concurrent = data["concurrent"] if "concurrent" in data else True

###############################################################################



###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file provides a simple logger internally used by the Coscine Python SDK.
The logger is capable of printing information to a specified file handle
including warnings and debug information.
A ProgressBar and color handling is also provided.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import TextIO, List, Callable
from enum import Enum
import sys
import platform
import colorama
from tqdm.auto import tqdm
from .__about__ import __version__

###############################################################################
# Class definition
###############################################################################

class LogLevel(Enum):
	DEBUG = 0
	WARN = 1
	NET = 2

###############################################################################
# Class definition
###############################################################################

class Colors:
	"""
	A simple class to handle colors in the terminal.
	"""

	BLUE: str = ""
	YELLOW: str = ""
	WHITE: str = ""

	@staticmethod
	def enable(enable: bool) -> None:
		"""
		Enable colors in command line output.
		"""

		if enable:
			Colors.BLUE = colorama.Fore.BLUE
			Colors.WHITE = colorama.Fore.WHITE
			Colors.YELLOW = colorama.Fore.YELLOW
		else:
			Colors.BLUE = Colors.WHITE = Colors.YELLOW = ""

###############################################################################
# Class definition
###############################################################################

class Logger:
	"""
	A simple logger class to handle standard output of the Coscine Python SDK.

	Attributes
	----------
	_stdout : TextIO
		The output file descriptor
	enabled : bool
		Enables/Disables command line output
	_loglevels : list
		A list of LogLevels to control the type of output of the logger.
	"""

	_stdout: TextIO
	enabled: bool
	_loglevels: List[LogLevel]

###############################################################################

	def __init__(self, enabled: bool, stdout: TextIO = sys.stdout,
						loglevels: List[LogLevel] = [LogLevel.WARN],
										colors: bool = True) -> None:
		"""
		Initializes an instance of the logger class.

		Parameters
		-----------
		enabled : bool
			Controls whether command line output is enabled.
		stdout : TextIO
			Overwrites the standard output file descriptor. Set to a file
			for logging to file.
		loglevels : List[LogLevel]
			Controls the type of output the logger makes.
		colors : bool
			Enables/Disables colored output. Should be
			disabled for logging to file.
		"""

		self._stdout = stdout
		self.enabled = enabled
		self.set_loglevels(loglevels)
		if colors:
			# Colorama overwrites sys.stdout, so special care is needed
			if self._stdout == sys.stdout:
				colorama.init()
				self._stdout = sys.stdout
				Colors.enable(True)

###############################################################################

	def banner(self) -> None:
		"""
		Prints an ASCII art banner of Coscine including
		the current SDK version to stdout.
		"""

		BANNER : str = f"""{Colors.BLUE} \
                    _             
                    (_)            
   ___ ___  ___  ___ _ _ __   ___  
  / __/ _ \/ __|/ __| | '_ \ / _ \ 
 | (_| (_) \__ \ (__| | | | |  __/ 
  \___\___/|___/\___|_|_| |_|\___| {Colors.WHITE}
____________________________________

    Coscine Python SDK {Colors.YELLOW}{__version__}{Colors.WHITE}
____________________________________
"""
		self.log(BANNER)

###############################################################################

	def set_loglevels(self, loglevels: List[LogLevel]) -> None:
		"""
		Set the loglevels of the logger to control what gets printed
		to the specified stdout file handle.

		Parameters
		----------
		loglevels : List[LogLevel]
			A list of LogLevel values
		
		Raises
		------
		ValueError
			In case the specified loglevels list contains an invalid loglevel.
		"""

		# Verify we got a valid loglevel
		for level in loglevels:
			if level not in LogLevel:
				raise ValueError(f"Specified LogLevel {level} is not \
												a valid loglevel!")

		# Apply loglevels
		self._loglevels = loglevels

###############################################################################

	def log(self, msg: str) -> None:
		"""
		If logging is enabled a message is printed to the specified
		stdout file handle.

		Parameters
		----------
		msg : str
			The message as a string.
		"""

		if self.enabled:
			print(msg, file=self._stdout)

###############################################################################

	def net(self, msg: str) -> None:
		"""
		Network traffic is printed to the specified stdout file handle if
		the loglevels include the LogLevel.NET level and logging is enabled.

		Parameters
		----------
		msg : str
			The message as a string.
		"""

		if LogLevel.NET in self._loglevels or LogLevel.DEBUG in self._loglevels:
			self.log("[NET] " + msg)

###############################################################################

	def warn(self, msg: str) -> None:
		"""
		A warning is printed to the specified stdout file handle if
		the loglevels include the LogLevel.WARN level and logging is enabled.

		Parameters
		----------
		msg : str
			The warning message as a string.
		"""

		if LogLevel.WARN in self._loglevels or LogLevel.DEBUG in self._loglevels:
			self.log("[WARNING] " + msg)

###############################################################################

	def debug(self, msg: str) -> None:
		"""
		A debug message is printed to the specified stdout file handle if
		the loglevels include the LogLevel.DEBUG level and logging is enabled.

		Parameters
		----------
		msg : str
			The debug message as a string.
		"""

		self.log("[DBG] " + msg)

###############################################################################

	@staticmethod
	def sysinfo() -> str:
		"""
		Constructs system information for better bug reports.

		Returns
		-------
		str
			Multiline string containing system and python information.
		"""

		info = \
		"""\
		Platform: %s
		Machine: %s
		Processor: %s
		Python compiler: %s
		Python branch: %s
		Python implementation: %s
		Python revision: %s
		Python version: %s
		""" % ( platform.platform(),
				platform.machine(),
				platform.processor(),
				platform.python_compiler(),
				platform.python_branch(),
				platform.python_implementation(),
				platform.python_revision(),
				platform.python_version()
			)
		return info.replace("\t", " ")

###############################################################################
# Class definition
###############################################################################

class ProgressBar:
	"""
	The ProgressBar class is a simple wrapper around tqdm
	progress bars. It is used in download/upload methods and provides the
	benefit of remembering state information and printing only when in
	verbose mode.

	Attributes
	----------
	logger : Logger
		Coscine Python SDK logger handle.
	bar : tqdm.tqdm
		tqdm Progress Bar instance.
	filesize : int
		Filesize in bytes.
	n : int
		Number of bytes read.
	callback : Callable[int]
		Callback function to call on update.
	"""

	logger: Logger
	bar: tqdm
	filesize: int
	n: int
	callback: Callable[[int]]

###############################################################################

	def __init__(self, logger: Logger, filesize: int, key: str, \
				mode: str, callback: Callable[[int]] = None) -> None:
		"""
		Initializes a state-aware tqdm ProgressBar.

		client : Client
			Coscine python SDK client handle
		filesize : int
			Size of the file object in bytes
		key : str
			key/filename of the file object
		mode : str
			'UP' or 'DOWN' referring to upload and download
		callafter : function(chunksize: int)
			 callback function to call after each update
		"""

		MODES = ("UP", "DOWN")

		if mode not in MODES:
			raise ValueError("Invalid value for argument 'mode'! " \
							"Possible values are %s." % str(MODES))

		self.logger = logger
		self.callback = callback
		self.filesize = filesize
		self.n = 0
		if self.logger.enabled:
			self.bar = tqdm(total=filesize, unit="B", unit_scale=True,
												desc="%s %s" % (mode, key))

###############################################################################

	def update(self, chunksize: int) -> None:
		"""
		Updates the progress bar with respect to the consumed chunksize.
		If a callafter function has been provided to the Constructor, it is
		called during the update.
		"""

		self.n += chunksize
		if self.logger.enabled:
			self.bar.update(chunksize)
			self.bar.refresh()
		if self.callback:
			self.callback(chunksize)

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file defines default and constant data internally used by
multiple modules to avoid redefinitions.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import List
from appdirs import AppDirs
from .__about__ import __version__

###############################################################################

# The time format for parsing Coscine date strings with datetime
TIMEFORMAT: str = "%Y-%m-%dT%H:%M:%S"

# Build default application directory paths
APPDIR: AppDirs = AppDirs("coscine-python-sdk", "Coscine")

# The name for the persistent cache file storage
CACHEFILE: str = f"{__version__}-cache.json"

# The languages supported by the Coscine Python SDK
LANGUAGES: List[str] = ["de", "en"]

# Supported formats for printing forms
TABLE_FORMATS: List[str] = ["str", "csv", "json", "html"]

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file defines the project object for the representation of
Coscine projects. It provides a simple interface to interact with Coscine
projects from python.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, List
if TYPE_CHECKING:
	from .client import Client
from .exceptions import *
from .resource import Resource, ResourceForm
from .form import InputForm
from .defaults import TIMEFORMAT
from prettytable.prettytable import PrettyTable
from datetime import datetime
from urllib.parse import urljoin, urlparse
import json
import os

###############################################################################
# Class definition
###############################################################################

class ProjectForm(InputForm):
	"""
	"""

	parent: Project

###############################################################################

	def __init__(self, client: Client, parent: Project = None) -> None:
		self._items = client.vocabularies.builtin("project")
		self.parent = parent
		super().__init__(client)
		vocabularies = {
			"disciplines": client.vocabularies.disciplines(True),
			"organizations": client.vocabularies.organizations(True),
			"visibility": client.vocabularies.visibility(True)
		}
		for item in self._items:
			if item.vocabulary:
				self._vocabularies[item.path] = vocabularies[item.path]

###############################################################################

	def _parse_organizations(self, data):
		organizations = []
		for value in data:
			organization = {
				#"displayName": urljoin(value["displayName"], urlparse(value["displayName"]).path),
				"url": urljoin(value["url"], urlparse(value["url"]).path)
			}
			organizations.append(organization)
			# organization vocabualry only contains 1st 100 entries
			# can only filter, not get full
		return organizations

###############################################################################

	def parse(self, data: dict) -> None:
		IGNORE = ["id", "slug", "parentId"]
		if data is None:
			return
		for path, value in data.items():
			if path in IGNORE: continue
			if path == "organizations":
				value = self._parse_organizations(value)
			key = self.name_of(path)
			self.set_value(key, value, True)

###############################################################################

	def generate(self) -> dict:
		"""
		"""

		metadata = {}
		missing = []

		for key, values in self.items(True):
			if not values:
				if self.is_required(key):
					missing.append(key)
				continue

			properties = self.properties(key)
			path = properties.path
			metadata[path] = values if properties.maxCount > 1 else values[0]

		if missing:
			raise ValueError(missing)

		if self.parent:
			metadata["ParentId"] = self.parent.id
		
		return metadata

###############################################################################
# Class definition
###############################################################################

class Project:
	"""
	Python representation of a Coscine Project
	"""

	client: Client
	data: dict
	parent: Project

	id: str
	name: str
	displayName: str
	description: str
	principleInvestigators: str
	startDate: datetime
	endDate: datetime
	disciplines: List[str]
	organizations: List[str]
	visibility: str

	###############################################################################

	def __init__(self, client: Client, data: dict,
					parent: Project = None) -> None:
		"""
		Initializes a Coscine project object.

		Parameters
		----------
		client : Client
			Coscine client handle
		data : dict
			Project data received from Coscine.
		parent : Project
			Optional parent project.
		"""

		self.client = client
		self.data = data
		self.parent = parent

###############################################################################

	@property
	def id(self) -> str: return self.data["id"]

	@property
	def name(self) -> str: return self.data["projectName"]

	@property
	def displayName(self) -> str: return self.data["displayName"]

	@property
	def description(self) -> str: return self.data["description"]

	@property
	def principleInvestigators(self) -> str:
		return self.data["principleInvestigators"]

	@property
	def startDate(self) -> datetime:
		return datetime.strptime(self.data["startDate"], TIMEFORMAT)

	@property
	def endDate(self) -> datetime:
		return datetime.strptime(self.data["endDate"], TIMEFORMAT)

	@property
	def disciplines(self) -> List[str]:
		lang = {
			"en": "displayNameEn",
			"de": "displayNameDe"
		}[self.client.language]
		return [k[lang] for k in self.data["disciplines"]]
	
	@property
	def organizations(self) -> List[str]:
		return[k["displayName"] for k in self.data["organizations"]]
	
	@property
	def visibility(self) -> str: return self.data["visibility"]["displayName"]

###############################################################################

	def __str__(self) -> str:
		table = PrettyTable(("Property", "Value"))
		rows = [
			("ID", self.id),
			("Name", self.name),
			("Display Name", self.displayName),
			("Principle Investigators", self.principleInvestigators),
			("Disciplines", "\n".join(self.disciplines)),
			("Organizations", "\n".join(self.organizations)),
			("Start Date", self.startDate),
			("End Date", self.endDate),
			("Visibility", self.visibility)
		]
		table.max_width["Value"] = 50
		table.add_rows(rows)
		return table.get_string(title = "Project [%s]" % self.displayName)

###############################################################################

	def subprojects(self, **kwargs) -> List[Project]:
		"""
		Retrieves a list of subprojects of the current project matching a set
		of specified filters

		Parameters
		----------
		kwargs
			key-value filters for subprojects (e.g. 'Name' = 'My Project').

		Returns
		-------
		list of Projects
		"""

		uri = self.client.uri("Project", "SubProject", self.id)
		projects = self.client.get(uri).json()
		filter = []
		for data in projects:
			match = True
			for key, value in kwargs.items():
				if data[key] != value:
					match = False
					break
			if match:
				filter.append(Project(self.client, data, self))
		return filter

###############################################################################

	def delete(self) -> None:
		"""
		Deletes the project on the Coscine servers.
		"""

		uri = self.client.uri("Project", "Project", self.id)
		self.client.delete(uri)

###############################################################################

	def resources(self, **kwargs) -> List[Resource]:
		"""
		Retrieves a list of Resources of the current project matching a set
		of specified filters.

		Parameters
		----------
		kwargs
			key-value filters for resources (e.g. 'Name' = 'My Resource').
		
		Returns
		-------
		list[Resource]
			list of resources matching the supplied filter.
		"""

		uri = self.client.uri("Project", "Project", self.id, "resources")
		resources = []
		for it in self.client.get(uri).json():
			for key, value in kwargs.items():
				if it[key] != value:
					break
			else:
				resources.append(Resource(self, it))
		return resources

###############################################################################

	def resource(self, displayName: str = None, **kwargs) -> Resource:
		"""
		Retrieves a certain resource of the current project matching a set
		of specified filters or identified by its displayName.

		Parameters
		----------
		displayName : str
			The display name of the resource.
		kwargs
			key-value filters for resources (e.g. 'Name' = 'My Resource').
		
		Returns
		--------
		Resource or None
		"""
		
		if displayName:
			kwargs["displayName"] = displayName
		elif not kwargs:
			raise ValueError("")

		resources = self.resources(**kwargs)
		if len(resources) == 1:
			return resources[0]
		elif len(resources) == 0:
			return None
		else:
			raise ValueError("Found more than 1 resource matching "\
										"the specified criteria!")

###############################################################################

	def download(self, path: str = "./", metadata: bool = False) -> None:
		"""
		Downloads the project to the location referenced by 'path'.

		Parameters
		----------
		path : str
			Download location on the harddrive
			Default: current directory './'
		metadata : bool, default: False
			If enabled, project metadata is downloaded and put in
			a hidden file '.metadata.json'.
		"""

		path = os.path.join(path, self.displayName)
		if not os.path.isdir(path):
			os.mkdir(path)
		for resource in self.resources():
			resource.download(path=path, metadata=metadata)
		if metadata:
			data = json.dumps(self.data, indent=4)
			fd = open(os.path.join(path, ".metadata.json"), "w")
			fd.write(data)
			fd.close()

###############################################################################

	def members(self) -> List[ProjectMember]:
		"""
		Retrieves a list of all members of the current project

		Returns
		--------
		list[ProjectMember]
			List of project members as ProjectMember objects.
		"""

		uri = self.client.uri("Project", "ProjectRole", self.id)
		data = self.client.get(uri).json()
		members = [ProjectMember(self, m) for m in data]
		return members

###############################################################################

	def invite(self, email: str, role: str = "Member") -> None:
		"""
		Invites a person to a project via their email address

		Parameters
		----------
		email : str
			The email address to send the invite to
		role : str, "Member" or "Owner", default: "Member"
			The role for the new project member
		"""

		if role not in ProjectMember.ROLES:
			raise ValueError("Invalid role '%s'." % role)

		uri = self.client.uri("Project", "Project", "invitation")
		data = {
			"projectId": self.data["id"],
			"role": ProjectMember.ROLES[role],
			"email": email
		}

		try:
			self.client.log("Inviting [%s] as [%s] to project [%s]." % 
												(email, role, self.id))
			self.client.post(uri, data = data)
		except ServerError:
			self.client.log("User [%s] has invite pending." % email)

###############################################################################

	def add_member(self, member: ProjectMember, role: str = "Member"):
		"""
		Adds a project member of another project to the current project.

		Parameters
		----------
		member : ProjectMember
			Member of another Coscine project
		role : str, "Member" or "Owner", default: "Member"
		"""

		if role not in ProjectMember.ROLES:
			raise ValueError("Invalid role!")

		data = member.data
		data["projectId"] = self.id
		data["role"]["displayName"] = role
		data["role"]["id"] = ProjectMember.ROLES[role]
		uri = self.client.uri("Project", "ProjectRole")
		self.client.post(uri, data = data)

###############################################################################

	def form(self) -> ProjectForm:
		"""
		"""

		form = ProjectForm(self.client)
		form.parse(self.data)
		return form

###############################################################################

	def update(self, form: ProjectForm) -> dict:
		"""
		Updates a project using the given ProjectForm

		Parameters
		----------
		form : ProjectForm
			ProjectForm containing updated data.
		"""

		if type(form) is ProjectForm:
			form = form.generate()
		uri = self.client.uri("Project", "Project", self.id)
		response = self.client.post(uri, data = form)
		if response.ok: self.data = response.json()
		return response

###############################################################################

	def create_resource(self, form: ResourceForm) -> Resource:
		"""
		Creates a resource within the current project using the supplied
		resource form.

		Parameters
		----------
		resourceForm : ResourceForm
			Form to generate the resource with.
		metadataPreset : MetadataPresetForm
			optional application profile configuration.
		"""
		if type(form) is ResourceForm:
			form = form.generate()
		uri = self.client.uri("Resources", "Resource", "Project", self.id)
		return Resource(self, self.client.post(uri, data = form).json())

###############################################################################
# Class definition
###############################################################################

class ProjectMember:

	"""
	Initializes a project member for a given project.

	Parameters
	----------
	project : Project
		Coscine python SDK project handle.
	data: dict
		User data as dict, retrieved via 
		client.uri("Project", "ProjectRole", self.id).
	"""

	ROLES = {
		"Owner": "be294c5e-4e42-49b3-bec4-4b15f49df9a5",
		"Member": "508b6d4e-c6ac-4aa5-8a8d-caa31dd39527"
	}

	name: str
	email: str
	id: str
	role: str
	data: dict
	client: Client
	project: Project

###############################################################################

	def __init__(self, project: Project, data: dict) -> None:
		"""
		"""

		self.project = project
		self.client = self.project.client
		self.data = data
		self.name = data["user"]["displayName"]
		self.email = data["user"]["emailAddress"]
		self.id = data["user"]["id"]
		self.role = data["role"]["displayName"]

###############################################################################

	def set_role(self, role: str) -> None:

		"""
		Sets the role of a project member

		Parameters
		----------
		role : str
			The new role of the member ('Owner' or 'Member').
		"""

		ROLES = {
			"Owner": "be294c5e-4e42-49b3-bec4-4b15f49df9a5",
			"Member": "508b6d4e-c6ac-4aa5-8a8d-caa31dd39527"
		}

		if role not in ROLES:
			raise ValueError("Invalid role '%s'." % role)

		uri = self.client.uri("Project", "ProjectRole")
		self.data["role"]["id"] = ROLES[role]
		self.data["role"]["displayName"] = role
		self.client.post(uri, data = self.data)

###############################################################################

	def remove(self) -> None:
		"""
		Removes a project member from their associated project.
		"""

		uri = self.client.uri("Project", "ProjectRole", "project", \
			self.project.id, "user", self.id, "role", self.data["role"]["id"])
		self.client.delete(uri)

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file implements a basic persistent cache for nonvolatile Coscine
data such as metadata vocabularies.
The data is automatically refreshed every once in a while.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from .__about__ import __version__
from .defaults import APPDIR, CACHEFILE, TIMEFORMAT
import atexit
import json
import os
from json.decoder import JSONDecodeError
from datetime import datetime, timedelta

###############################################################################
# Class
###############################################################################

class Cache:
	"""
	A basic persistent cache for temporary storage of nonvolatile data.
	Data is automatically refreshed when it exceeds a certain age.
	Data is loaded from & stored on disk if persistency option is enabled.

	Attributes
	-----------
	_cache : dict
		A simple dictionary to store the cached data.
	_persistent : bool
		Controls cache persistence - if enabled the cache is saved to
		a file and restored upon the next session.
	"""

	_cache: dict
	_persistent: bool

###############################################################################

	def __init__(self, persistent: bool = True) -> None:
		"""
		Initializes a Cache instance and attempts to load a previous
		cache copy from a file if the $persistent option is set.

		Parameters
		----------
		persistent : bool
			Enable to save the cache in a file upon program exit and
			to restore it in the next session.
		"""

		self._persistent = persistent
		self.load()

###############################################################################

	def save(self) -> None:
		"""
		Saves the current session cache into a file if
		the $persistent option is set (see __init__()).
		Any cachefile existing prior is overwritten.
		"""

		if self._persistent and self._cache:
			path = APPDIR.user_cache_dir
			if not os.path.exists(path):
				os.makedirs(path, exist_ok=True)
			filepath = os.path.join(path, CACHEFILE)
			fd = open(filepath, "w")
			fd.write(json.dumps(self._cache))
			fd.close()

###############################################################################

	def load(self) -> None:
		"""
		Loads a previously stored session cache for use with the
		current session. If no previous session is found, the cache
		is intialized as an empty dict.
		"""

		if self._persistent:
			atexit.register(self.save)
			try:
				path = APPDIR.user_cache_dir
				filepath = os.path.join(path, CACHEFILE)
				fd = open(filepath, "r")
				self._cache = json.loads(fd.read())
				fd.close()
			except (FileNotFoundError, JSONDecodeError):
				self._cache = {}
		else:
			self._cache = {}

###############################################################################

	def get(self, key: str) -> str:
		"""
		Attempts to read a cached dataset from the cache via a key.

		Parameters
		----------
		key : str
			The key used to save data in the cache (e.g. a URL)
		
		Returns
		--------
		str
			The cached data for the given key
		None
			If the key is not present in the cache or if the data is
			likely to be outdated
		"""

		if key in self._cache:
			lastTime = datetime.strptime(self._cache[key]["time"], TIMEFORMAT)
			if ((datetime.now() - lastTime) < timedelta(days=24)):
				return self._cache[key]["data"]
		return None

###############################################################################

	def set(self, key: str, data) -> None:
		"""
		Sets data inside the cache via a key. If that key is already
		present any data referenced by it is overwritten.

		Parameters
		----------
		key : str
			The key used for saving data in the cache (e.g. a URL)
		"""

		self._cache[key] = {
			"time": datetime.now().strftime(TIMEFORMAT),
			"data": data
		}

###############################################################################

	def clear(self) -> None:
		"""
		Clears the cache
		"""

		self._cache.clear()

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file implements various classes for querying, parsing and interacting
with data inside Coscine vocabularies.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, List, Tuple
if TYPE_CHECKING:
	from .client import Client
import os
import json
import pkgutil
import urllib.parse
from .defaults import LANGUAGES
from .form import FormField
from .graph import ApplicationProfile

###############################################################################
# Class
###############################################################################

class Vocabulary:
	"""
	A wrapper around Coscine vocabularies. Vocabularies from Coscine
	do not necessarily all follow the same format, so this class takes
	care of abstracting the Vocabulary interface.
	The internal language of the Vocabulary changes automatically based
	on the Client language.

	Attributes
	-----------
	raw : dict
		Contains the raw dictionary data that is served by Coscine. Parsing
		for language is not taken into account.
	content : dict
		Contains the raw dictionary data that is served by Coscine
		for the language preset set in coscine.Client .
	"""

	client: Client
	_data: dict

	@property
	def raw(self) -> dict: return self._data

	@property
	def content(self) -> dict:
		data = self.raw
		languages = self.languages()
		# Select appropriate language branch
		if self.client.language in languages:
			data = data[self.client.language]
		elif len(languages) > 0:
			data = data[languages[0]]
		return data

###############################################################################

	def __init__(self, client: Client, data: dict) -> None:
		self.client = client
		self._data = data

###############################################################################

	def __str__(self) -> str: return json.dumps(self.raw, indent=4)

###############################################################################

	def keys(self) -> List[str]:
		return [entry["name"] for entry in self.content]

###############################################################################

	def values(self) -> List[object]:
		return [entry["value"] for entry in self.content]

###############################################################################

	def items(self) -> List[Tuple[str, object]]:
		return zip(self.keys(), self.values())

###############################################################################

	def languages(self) -> List[str]:
		"""
		Returns the languages supported by the dictionary.
		Some dictionaries may not contain a language setting and thus
		are valid for all languages. In that case an empty list is returned.
		"""

		languages = []
		for lang in LANGUAGES:
			if lang in self.raw:
				languages.append(lang)
		return languages

###############################################################################

	def contains(self, key: object, reverse: bool = False) -> bool:
		"""
		Determines whether the vocabulary contains the given key.

		Parameters
		----------
		key : object (str, list or dict)
			The key to look for in the vocabulary.
		reverse : bool
			If set to true, key is assumed to be a value inside of the
			vocabulary and thus all values are searched for key.
		"""

		return self.lookup(key, reverse) is not None

###############################################################################

	def lookup(self, key: object, reverse: bool = False) -> object:
		"""
		Looks up the counterpart of $key.

		Parameters
		----------
		key : object (str, list or dict)
			The key to look for in the vocabulary.
		reverse : bool
			If set to true, key is assumed to be a value inside of the
			vocabulary and thus all values are searched for key.
		"""

		if reverse: return self._lookup_value(key)
		else: return self._lookup_key(key)

###############################################################################

	def _lookup_key(self, key: object) -> object:
		"""
		Returns the value referenced by key
		"""

		for entry in self.content:
			if entry["name"] == key: return entry["value"]
		return None

###############################################################################

	def _lookup_value(self, value: object) -> object:
		"""
		Returns the key referenced by value
		"""

		if type(value) is list and len(value) == 1: value = value[0]
		for entry in self.content:
			if type(value) is str:
				if entry["value"] == value: return entry["name"]
			elif type(value) is dict:
				if self._is_subset(value, entry["value"]): return entry["name"]
		return None

###############################################################################

	@staticmethod
	def _is_subset(x: dict, y: dict) -> bool:
		"""
		Check whether x is a subset of y
		"""
		return x.items() <= y.items()

###############################################################################
# Class
###############################################################################

class VocabularyManager:
	"""
	The VocabularyManager takes care of loading local InputForms and
	querying vocabularies from the Coscine servers. Local InputForms
	are stored in src/data in json format.
	It makes use of the Cache class thus making it more efficient at
	querying vocabularies.
	"""

	client: Client
	_builtins: dict

###############################################################################

	def __init__(self, client: Client) -> None:
		"""
		Initializes the VocabularyManager with a client instance for
		talking to the Coscine servers.
		"""
		self.client = client
		self._builtins = {
			"project": json.loads(
				pkgutil.get_data(__name__, "data/project.json").decode("utf-8")
			),
			"resource": json.loads(
				pkgutil.get_data(__name__, "data/resource.json").decode("utf-8")
			)
		}
		for graph in self._builtins: self._builtins[graph] = [
			FormField(client, element) for element in self._builtins[graph]
		]

###############################################################################

	def builtin(self, name: str) -> List[FormField]:
		"""
		Returns a builtin InputForm as a list of FormFields.

		Parameters
		-----------
		name : str
			Name of the builtin vocabulary (filename in src/data
			without filetype).
		"""
		if name not in self._builtins:
			raise ValueError("Invalid builtin InputForm name!")
		return self._builtins[name]

###############################################################################

	def licenses(self, normalize: bool = True) -> Vocabulary:
		"""
		Returns a dictionary containing a list of licenses
		available in Coscine.
		"""

		uri = self.client.uri("Project", "License")
		data = self.client.static_request(uri)
		if normalize:
			normalized: dict = {"en": []}
			for entry in data:
				normalized["en"].append({
					"name": entry["displayName"],
					"value": entry
				})
			data = normalized
		return Vocabulary(self.client, data)

###############################################################################

	def resource_types(self, normalize: bool = True) -> Vocabulary:
		"""
		Retrieves a list of the available resource types in Coscine.
		Only enabled (currently usable) resource types are returned.
		"""

		uri = self.client.uri("Resources", "ResourceType", "types")
		data = self.client.static_request(uri)
		if normalize:
			normalized: dict = {"en": []}
			for entry in data:
				if not entry["isEnabled"]: continue
				normalized["en"].append({
					"name": entry["displayName"],
					"value": entry
				})
			data = normalized
		return Vocabulary(self.client, data)

###############################################################################

	def application_profiles(self, normalize: bool = True) -> Vocabulary:
		"""
		Returns a list of all available Coscine application profiles.
		"""

		uri = self.client.uri("Metadata", "Metadata", "profiles")
		data = self.client.static_request(uri)
		if normalize:
			normalized: dict = {"en": []}
			for entry in data:
				name = urllib.parse.urlparse(entry)[2]
				name = os.path.relpath(name, "/coscine/ap/")
				normalized["en"].append({
					"name": name,
					"value": entry
				})
			data = normalized
		return Vocabulary(self.client, data)

###############################################################################

	def application_profile(self, path: str) -> ApplicationProfile:
		"""
		Retrieves a specific Coscine application profile.

		Parameters
		-----------
		path : str
			Path/Url to the application profile.
			(e.g. Resource.data["applicationProfile"]))
		id : str, default: None
			Coscine resource ID.
		
		Returns
		-------
		ApplicationProfile
			A parsed application profile.
		"""

		uri = self.client.uri("Metadata", "Metadata", "profiles", path)
		data = self.client.static_request(uri)
		graph = json.dumps(data[0]["@graph"])
		return ApplicationProfile(graph)

###############################################################################

	def instance(self, link: str) -> Vocabulary:
		"""
		Retrieves a data instance. Instances are always normalized.

		Parameters
		-----------
		link : str
			link to the instance
		"""

		uri = self.client.uri("Metadata", "Metadata", "instances", link)
		instance = self.client.static_request(uri)
		return Vocabulary(self.client, instance)

###############################################################################

	def organizations(self, normalize: bool = True, filter: str = "") -> Vocabulary:
		"""
		Queries the organizations (e.g. 'RWTH Aachen University') available
		for selection in Coscine.
		"""

		if filter:
			raise NotImplementedError("Handling of filter argument not implemented!")
		else:
			uri = self.client.uri("Organization", "Organization", "-", "ror")
		data = self.client.static_request(uri)
		if normalize:
			normalized: dict = {"en": []}
			for entry in data["data"]:
				normalized["en"].append({
					"name": entry["displayName"],
					"value": entry
				})
			data = normalized
		return Vocabulary(self.client, data)

###############################################################################

	def visibility(self, normalize: bool = True) -> Vocabulary:
		"""
		Returns the key-value mapping of the Coscine project visibility field.

		Returns
		-------
		dict
			Key-value mapped Coscine project visibility field.
		"""

		uri = self.client.uri("Project", "Visibility")
		data = self.client.static_request(uri)
		if normalize:
			normalized: dict = {"en": []}
			for entry in data:
				normalized["en"].append({
					"name": entry["displayName"],
					"value": entry
				})
			data = normalized
		return Vocabulary(self.client, data)

###############################################################################

	def disciplines(self, normalize: bool = True) -> Vocabulary:
		"""
		Queries the scientific disciplines available for selection in Coscine.

		Returns
		-------
		dict
			Dictionary containing the disciplines.
		"""

		uri = self.client.uri("Project", "Discipline")
		data = self.client.static_request(uri)
		if normalize:
			normalized: dict = {"en": [], "de": []}
			for entry in data:
				normalized["en"].append({
					"name": entry["displayNameEn"],
					"value": entry
				})
				normalized["de"].append({
					"name": entry["displayNameDe"],
					"value": entry
				})
			data = normalized
		return Vocabulary(self.client, data)

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file defines the resource object for the representation of
Coscine resources. It provides an easy interface to interact with Coscine
resources from python.
"""

###############################################################################
# Dependencies
###############################################################################

from __future__ import annotations
from typing import TYPE_CHECKING, List, Callable
if TYPE_CHECKING:
	from .client import Client
	from .project import Project
from .exceptions import *
from .object import FileObject, MetadataForm
from .logger import ProgressBar
from .form import InputForm
from .graph import ApplicationProfile
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from prettytable.prettytable import PrettyTable
from concurrent.futures import ThreadPoolExecutor, wait
import json
import os

###############################################################################
# Class definition
###############################################################################

class S3info:
	"""
	Provides a simple pythonic interface to resource S3 access data
	"""

	read_access_key: str
	read_secret_key: str
	write_access_key: str
	write_secret_key: str
	endpoint: str
	bucket: str
	_data: dict

	def __init__(self, data: dict) -> None:
		self._data = data
		if "resourceTypeOption" in data and data["resourceTypeOption"]:
			opt: dict = data["resourceTypeOption"]
			self.read_access_key = opt.get("ReadAccessKey")
			self.read_secret_key = opt.get("ReadSecretKey")
			self.write_access_key = opt.get("WriteAccessKey")
			self.write_secret_key = opt.get("WriteSecretKey")
			self.endpoint = opt.get("Endpoint")
			self.bucket = opt.get("BucketName")

###############################################################################
# Class definition
###############################################################################

class ResourceForm(InputForm):
	"""
	"""

###############################################################################

	def __init__(self, client: Client) -> None:
		self._items = client.vocabularies.builtin("resource")
		super().__init__(client)
		vocabularies = {
			"type": client.vocabularies.resource_types(True),
			"applicationProfile": client.vocabularies.application_profiles(True),
			"license": client.vocabularies.licenses(True),
			"visibility": client.vocabularies.visibility(True),
			"disciplines": client.vocabularies.disciplines(True)
		}
		for item in self._items:
			if item.vocabulary:
				self._vocabularies[item.path] = vocabularies[item.path]

###############################################################################

	def parse(self, data: dict) -> None:
		IGNORE = ["id", "pid", "fixedValues", "creator", "archived"]
		if data is None:
			return
		for path, value in data.items():
			if path in IGNORE: continue
			key = self.name_of(path)
			self.set_value(key, value, True)

###############################################################################

	def generate(self) -> dict:
		metadata = {}

		# Collect missing required fields
		missing = []

		# Set metadata
		for key, values in self.items(True):
			
			if not values:
				if self.is_required(key):
					missing.append(key)
				continue

			properties = self.properties(key)
			metadata[properties.path] = values if len(values) > 1 else values[0]

		# Check for missing required fields
		if len(missing) > 0:
			if len(missing) == 1 and missing[0] == "resourceTypeOption" \
									and	metadata["type"] == "linked":
				pass
			else:
				raise ValueError(missing)

		return metadata

###############################################################################
# Class definition
###############################################################################

class Resource:
	"""
	Python representation of a Coscine Resource type.
	"""

	client: Client
	project: Project
	data: dict

	id: str
	pid: str
	name: str
	displayName: str
	type: str
	disciplines: List[str]
	profile: str
	archived: bool
	creator: str
	s3: S3info

###############################################################################

	def __init__(self, project: Project, data: dict) -> None:
		"""
		Initializes a Coscine resource object.

		Parameters
		----------
		project : Project
			Coscine project handle
		data : dict
			Resource data received from Coscine.
		"""

		self.project = project
		self.client = self.project.client
		self.data = data
		self.s3 = S3info(data)

###############################################################################

	@property
	def id(self) -> str: return self.data["id"]

	@property
	def pid(self) -> str: return self.data["pid"]

	@property
	def name(self) -> str: return self.data["resourceName"]

	@property
	def displayName(self) -> str: return self.data["displayName"]

	@property
	def license(self) -> str:
		return self.data["license"]["displayName"] \
			if self.data["license"] else ""

	@property
	def type(self) -> str: return self.data["type"]["displayName"]

	@property
	def disciplines(self) -> List[str]:
		lang = {
			"en": "displayNameEn",
			"de": "displayNameDe"
		}[self.client.language]
		return [k[lang] for k in self.data["disciplines"]]

	@property
	def profile(self) -> str: return self.data["applicationProfile"]

	@property
	def archived(self) -> bool: return bool(self.data["archived"])

	@property
	def creator(self) -> str: return self.data["creator"]

###############################################################################

	def __str__(self) -> str:
		table = PrettyTable(["Property", "Value"])
		rows = [
			("ID", self.id),
			("Resource Name", self.name),
			("Display Name", self.displayName),
			("PID", self.pid),
			("Type", self.type),
			("Disciplines", "\n".join(self.disciplines)),
			("License", self.license),
			("Application Profile", self.profile),
			("Archived", self.archived),
			("Creator", self.creator),
			("Project", self.project.displayName),
			("Project ID", self.project.id)
		]
		table.max_width["Value"] = 50
		table.add_rows(rows)
		return table.get_string(title = "Resource [%s]" % self.displayName)

###############################################################################

	def delete(self) -> None:
		"""
		Deletes the Coscine resource and all objects contained within it on
		the Coscine servers.
		"""

		uri = self.client.uri("Resources", "Resource", self.id)
		self.client.delete(uri)

###############################################################################

	def application_profile(self) -> ApplicationProfile:
		"""
		Returns the application profile of the resource
		"""

		return self.client.vocabularies.application_profile(self.profile)

###############################################################################

	def _download_concurrently(self, path: str = "./") -> None:
		with ThreadPoolExecutor(max_workers=4) as executor:
			files = self.objects()
			futures = [executor.submit(file.download, path) for file in files]
			wait(futures)
			for index, fut in enumerate(futures):
				try: fut.result()
				except CoscineException:
					self.client.logger.warn(
						f"Error downloading '{files[index].name}'.")

###############################################################################

	def download(self, path: str = "./", metadata: bool = False) -> None:
		"""
		Downloads the resource and all of its contents to the local harddrive.

		Parameters
		----------
		path : str, default: "./"
			Path to the local storage location.
		metadata : bool, default: False
			If enabled, resource metadata is downloaded and put in
			a hidden file '.metadata.json'.
		"""

		path = os.path.join(path, self.name)
		if not os.path.isdir(path):
			os.mkdir(path)
		if self.client.concurrent:
			self._download_concurrently(path)
		else:
			for file in self.objects():
				try: file.download(path)
				except CoscineException:
					self.client.logger.warn(f"Error downloading '{file.name}'.")
		if metadata:
			data = json.dumps(self.data, indent=4)
			with open(os.path.join(path, ".resource-metadata.json"), "w") as fd:
				fd.write(data)

###############################################################################

	def objects(self, path: str = None, **kwargs) -> List[FileObject]:
		"""
		Returns a list of Objects stored within the resource

		Parameters
		------------
		kwargs
			file-object filter arguments (e.g. 'Name' = 'testfile').

		Returns
		-------
		list[Object]
			List of Coscine file-like objects.
		"""

		objects = []
		uri = self.client.uri("Tree", "Tree", self.id)
		if path:
			args = {"path": path}
		else:
			args = None
		data = self.client.get(uri, params = args).json()
		fileStorage = data["data"]["fileStorage"]
		metadataStorage = data["data"]["metadataStorage"]
		for data in fileStorage:
			for key, value in kwargs.items():
				if data[key] != value:
					break
			else:
				objects.append(FileObject(self, data))
		return objects

###############################################################################

	def object(self, path: str, **kwargs) -> FileObject:
		"""
		Returns an Object stored within the resource

		Parameters
		------------
		displayName : str, default: None
			file-object display name (filename/key).
		kwargs
			Filter

		Returns
		-------
		FileObject
			Python representation of the file-object as an Object instance
		"""
		
		objects = self.objects(Path=path, **kwargs)
		if len(objects) == 1:
			return objects[0]
		elif len(objects) == 0:
			return None
		else:
			raise ValueError("Found more than 1 FileObject matching "\
											"the specified criteria!")

###############################################################################

	def upload(self, key: str, file, metadata = None, \
					callback: Callable[[int]] = None) -> None:
		"""
		Uploads a file-like object to a resource on the Coscine server

		Parameters
		----------
		key : str
			filename of the file-like object.
		file : object with read() attribute
				Either open file handle or local file location path.
		metadata : dict
			File metadata. For rds-s3 this is optional, but recommended.
		callback : Callable[int]
			Optional callback called during chunk uploads.
		"""

		if hasattr(file, "read"):
			fd = file
			filename = "MEM"
		elif type(file) is str:
			fd = open(file, "rb")
			filename = file
		else:
			raise TypeError("Argument `file` has unexpected type!")

		if metadata:
			if type(metadata) is MetadataForm:
				metadata = metadata.generate()
			uri = self.client.uri("Tree", "Tree", self.id, key)
			self.client.put(uri, data = metadata)

		uri = self.client.uri("Blob", "Blob", self.id, key)
		fields = {"files": (key, fd, "application/octect-stream")}
		encoder = MultipartEncoder(fields = fields)
		bar = ProgressBar(self.client.logger, encoder.len, filename, "UP", callback)
		monitor = MultipartEncoderMonitor(encoder, callback = \
				lambda monitor: bar.update(monitor.bytes_read - bar.n))
		headers = {"Content-Type": monitor.content_type}
		self.client.put(uri, data = monitor, headers = headers)

###############################################################################

	def set_archived(self, flag: bool) -> None:
		"""
		Set the archived flag of the resource to put it in read-only mode.
		Only the resource creator or project owner can do this.

		Parameters
		----------
		flag : bool
			Enable with True, Disable with False.
		"""

		uri = self.client.uri("Resources", "Resource", self.id, \
					"setReadonly?status=%s" % str(flag).lower())
		self.client.post(uri)
		self.data["archived"] = flag
		self.archived = flag

###############################################################################

	def form(self) -> ResourceForm:
		"""
		Returns a ResourceForm filled with the metadata of the current resource.

		Returns
		-------
		ResourceForm
		"""

		form = self.client.forms.resource()
		form.parse(self.data)
		return form

###############################################################################

	def update(self, form: ResourceForm) -> dict:
		"""
		Updates the metadata of the resource using the supplied ResourceForm.

		Parameters
		----------
		form : ResourceForm
			ResourceForm filled with updated values.
		"""

		if type(form) is ResourceForm:
			form = form.generate()
		elif type(form) is not dict:
			raise TypeError("")

		uri = self.client.uri("Resources", "Resource", self.id)
		response = self.client.post(uri, data = form)
		if response.ok: self.data = response.json()
		return response

###############################################################################

	def metadata(self, data: dict = None) -> MetadataForm:
		"""
		Creates a MetadataForm for this resource
		"""

		return self.client.forms.metadata(self.application_profile())

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file contains package metadata primarily intended for setup.py but
also accessed by various source files in src/.
"""

###############################################################################

# Package title/name as it would appear in PyPi
__title__ = "coscine"

# Current package version
# Do not set version to 1.0.0 before Coscine end of pilot phase
__version__ = "0.7.2"

# Short package description
__summary__ = "The Coscine Python SDK is a pythonic interface to " \
			  "the Coscine REST API."

# Package copyright owner
__author__ = "RWTH Aachen University"

# Coscine contact (Note: This is the official Coscine contact email!)
# Only for copyright claims, licensing issues or sourcecode hosting!
# Please do not direct bug reports or feature requests to that address!
__contact__ = "coscine@itc.rwth-aachen.de"

# Package license
__license__ = "MIT License"

# Package keywords/tags (must be given as a list!)
__keywords__ = ["Coscine", "RWTH Aachen", "Research Data Management"]

# Package dependencies (must be given as a list!)
__dependencies__ = [
	"rdflib",
	"requests",
	"requests-toolbelt",
	"tqdm",
	"colorama",
	"prettytable",
	"appdirs"
]

# Python version required
__pyver__ = ">=3.7"

# Project url (official sourcecode hosting)
__url__ = "https://git.rwth-aachen.de/coscine/community-features/" \
											"coscine-python-sdk/"

# Additional urls for e.g. bug reports and documentation
__project_urls__ = {
	"Issues":  __url__ + "-/issues",
	"Documentation": "https://coscine.pages.rwth-aachen.de/" \
					 "community-features/coscine-python-sdk/coscine/"
}

###############################################################################



###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This file defines all of the exceptions raised by the Coscine Python SDK.
The base exception class is called CoscineException. It directly inherits
its properties from the standard Python Exception class.
Although never used directly it provides the basis for more specific
exceptions, thus allowing you to catch any Coscine exception with just one
except statement:
try:
	something()
except coscine.CoscineException: # Gotta Catch 'Em All
	print("Oh Dios Mio!")
"""

###############################################################################
# Class definition
###############################################################################

class CoscineException(Exception):
	"""
	Coscine base Exception class.
	"""

###############################################################################

class ConnectionError(CoscineException):
	"""
	A ConnectionError is raised in case the client is not able to establish
	a connection with the Coscine servers.
	"""
	pass

###############################################################################

class AuthorizationError(CoscineException):
	"""
	AuthorizationErrors are thrown in case of an invalid
	or empty Coscine ApiToken.
	"""
	pass

###############################################################################

class PermissionError(CoscineException):
	"""
	A PermissionError is thrown in cases where the owner of
	the Coscine API token does not hold enough privileges to access
	a certain path or perform a certain action.
	"""
	pass

###############################################################################

class ClientError(CoscineException):
	"""
	An ambiguous error has been made or detected on the client side.
	"""
	pass

###############################################################################

class ServerError(CoscineException):
	"""
	An ambiguous error has been made or detected on the Coscine server.
	"""
	pass

###############################################################################


###############################################################################
# Coscine Python SDK
# Copyright (c) 2018-2022 RWTH Aachen University
# Licensed under the terms of the MIT License
# #############################################################################
# Coscine, short for Collaborative Scientific Integration Environment is
# a platform for research data management (RDM).
# For more information on Coscine visit https://www.coscine.de/.
#
# Please note that this python module is open source software primarily
# developed and maintained by the scientific community. It is not
# an official service that RWTH Aachen provides support for.
###############################################################################

###############################################################################
# File description
###############################################################################

"""
This setup script builds and installs the Coscine Python SDK package.
For a local installation use:
cd ./coscine-python-sdk
py -m pip install .
"""

###############################################################################

import setuptools

###############################################################################

# Serves as a container for package metadata dynamically
# loaded from ./src/coscine/about.py
about = {}

# Read package metadata into $about
with open("src/coscine/__about__.py") as fd:
	exec(fd.read(), about)

# Use README.md as package description for display on PyPi
# and similar sourcecode hosting platforms.
# Prepends local links with the repository url to ensure display of
# images and other media on external platforms such as PyPi.
# Local links in README.md are marked with the prefix './//'.
with open("README.md", "r", encoding="utf-8") as fd:
	readme = fd.read()
	description = readme.replace("./data", about["__url__"]+"-/raw/master/data")

###############################################################################

setuptools.setup(
	name = about["__title__"],
	version = about["__version__"],
	description = about["__summary__"],
	long_description = description,
	long_description_content_type = "text/markdown",
	author = about["__author__"],
	author_email = about["__contact__"],
	license = about["__license__"],
	packages = setuptools.find_packages(where="src"),
	keywords = about["__keywords__"],
	package_data = {"coscine": ["data/*"]},
	install_requires = about["__dependencies__"],
	url = about["__url__"],
	project_urls = about["__project_urls__"],
	classifiers = [
		"Programming Language :: Python :: 3.7",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		"Intended Audience :: Developers"
	],
	package_dir = {"": "src"},
	python_requires = about["__pyver__"]
)

###############################################################################
